/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.extension.command.monitoring;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingArgumentException;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.UnrecognizedOptionException;
import org.ow2.petals.cli.api.command.Command;
import org.ow2.petals.cli.api.command.CommandUtil;
import org.ow2.petals.cli.api.command.exception.CommandBadArgumentNumberException;
import org.ow2.petals.cli.api.command.exception.CommandException;
import org.ow2.petals.cli.api.command.exception.CommandInvalidArgumentException;
import org.ow2.petals.cli.api.command.exception.CommandInvalidException;
import org.ow2.petals.cli.api.command.exception.CommandMissingArgumentException;
import org.ow2.petals.cli.api.command.exception.CommandMissingOptionsException;
import org.ow2.petals.cli.api.command.exception.ConnectionRequiredException;
import org.ow2.petals.cli.api.shell.Shell;
import org.ow2.petals.cli.extension.command.monitoring.exception.SubscriptionAlreadyExistsException;
import org.ow2.petals.cli.extension.command.monitoring.exception.SubscriptionObjectNotSubscribedException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.MonitoringObject;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.MonitoringObjectRegistration;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.MonitoringObjectRegistror;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.MonitoringSubFunction;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.DuplicatedMonitoringObjectException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.DuplicatedMonitoringSubFunctionException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.InvalidMonitoringObjectNameException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.InvalidMonitoringSubFunctionNameException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.MonitoringSubFunctionBadArgumentNumberException;
import org.ow2.petals.cli.extension.command.monitoring.mo.api.exception.MonitoringSubFunctionException;
import org.ow2.petals.cli.extension.command.monitoring.so.api.EmittableDefect;
import org.ow2.petals.cli.extension.command.monitoring.so.api.SubscriptionObject;
import org.ow2.petals.cli.extension.command.monitoring.so.api.SubscriptionObjectRegistration;
import org.ow2.petals.cli.extension.command.monitoring.so.api.SubscriptionObjectRegistror;
import org.ow2.petals.cli.extension.command.monitoring.so.api.exception.DuplicatedEmittableDefectException;
import org.ow2.petals.cli.extension.command.monitoring.so.api.exception.DuplicatedSubscriptionObjectException;
import org.ow2.petals.cli.extension.command.monitoring.so.api.exception.InvalidEmittableDefectNameException;
import org.ow2.petals.cli.extension.command.monitoring.so.api.exception.InvalidSubscriptionObjectNameException;
import org.ow2.petals.cli.extension.command.monitoring.so.api.exception.SubscriptionObjectException;

import jline.console.completer.Completer;
import jline.console.completer.StringsCompleter;

/**
 * The command 'monitoring'
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class Monitoring implements Command, MonitoringObjectRegistror, SubscriptionObjectRegistror {

    private static final Logger LOG = Logger.getLogger(Monitoring.class.getName());

    /**
     * The command name
     */
    public static final String NAME = "monitoring";

    /**
     * The command description
     */
    public static final String DESCRIPTION = "Get monitoring information, or subscribe to defect";

    /**
     * The option associated to the object key providing expected monitoring
     * information.
     */
    public static final String OBJECT_SHORT_OPTION = "o";

    private static final String OBJECT_LONG_OPTION = "object";

    private static final String OBJECT_ARG_NAME = "object";

    private static final Option OBJECT_OPTION = Option.builder(OBJECT_SHORT_OPTION).longOpt(OBJECT_LONG_OPTION)
            .required(false).numberOfArgs(1).argName(OBJECT_ARG_NAME)
            .desc("Component key providing expected monitoring information.").build();

    /**
     * The option associated to the sub-function key providing expected
     * monitoring information.
     */
    public static final String SUBFONCTION_SHORT_OPTION = "f";

    private static final String SUBFONCTION_LONG_OPTION = "sub-fonction";

    private static final String SUBFONCTION_ARG_NAME = "sub-fonction";

    private static final Option SUBFONCTION_OPTION = Option.builder(SUBFONCTION_SHORT_OPTION).numberOfArgs(1)
            .argName(SUBFONCTION_ARG_NAME).required(false).longOpt(SUBFONCTION_LONG_OPTION)
            .desc("Key identifying the expected monitoring information.").build();

    /**
     * The option associated to the object key providing subscription to defect
     */
    public static final String SUBSCRIPTION_SHORT_OPTION = "s";

    private static final String SUBSCRIPTION_LONG_OPTION = "subscribe";

    private static final String SUBSCRIPTION_ARG_NAME = "subscription object";

    private static final Option SUBSCRIPTION_OPTION = Option.builder(SUBSCRIPTION_SHORT_OPTION)
            .longOpt(SUBSCRIPTION_LONG_OPTION).required(false).numberOfArgs(1).argName(SUBSCRIPTION_ARG_NAME)
            .desc("Component key providing subscription to defect.").build();

    /**
     * The option associated to the object key providing unsubscription of
     * defect
     */
    public static final String UNSUBSCRIPTION_SHORT_OPTION = "u";

    private static final String UNSUBSCRIPTION_LONG_OPTION = "unsubscribe";

    private static final String UNSUBSCRIPTION_ARG_NAME = "subscription object";

    private static final Option UNSUBSCRIPTION_OPTION = Option.builder(UNSUBSCRIPTION_SHORT_OPTION)
            .longOpt(UNSUBSCRIPTION_LONG_OPTION).required(false).numberOfArgs(1).argName(UNSUBSCRIPTION_ARG_NAME)
            .desc("Component key providing unsubscription of defect.").build();

    /**
     * The option associated to the help of a sub-function
     */
    public static final String HELP_SHORT_OPTION = "h";

    private static final String HELP_LONG_OPTION = "help";

    private static final Option HELP_OPTION = Option.builder(HELP_SHORT_OPTION).longOpt(HELP_LONG_OPTION)
            .desc("Help on a sub-function of a monitoring object").build();

    private final Map<String, Completer> completers = new HashMap<>();

    /**
     * The command options
     */
    private Options options;

    /**
     * The shell this command is bound to
     */
    private Shell shell;

    /**
     * Monitoring object registered
     */
    private Map<String, MonitoringObject> monitoringObjects = new HashMap<>();

    /**
     * Subscription object registered
     */
    private Map<String, SubscriptionObject> subscriptionObjects = new HashMap<>();

    /**
     * The list of subscription objects for which an subscription exists
     */
    private List<String> currentSubscriptions = new LinkedList<>();

    public Monitoring() {
        super();
        this.createOptions();

        // Load monitoring objects
        MonitoringObjectRegistration.registers(this);

        // Load subscription objects
        SubscriptionObjectRegistration.registers(this);

        // Option completers
        final Completer monitoringObjectCompleter = new StringsCompleter(
                this.monitoringObjects.keySet());
        this.completers.put(OBJECT_SHORT_OPTION, monitoringObjectCompleter);
        final Completer subscriptionObjectCompleter = new StringsCompleter(
                this.subscriptionObjects.keySet());
        this.completers.put(SUBSCRIPTION_SHORT_OPTION, subscriptionObjectCompleter);
        this.completers.put(UNSUBSCRIPTION_SHORT_OPTION, subscriptionObjectCompleter);
    }

    @Override
    public void registersMonitoringObjects(final MonitoringObject monitoringObject)
            throws DuplicatedMonitoringObjectException, DuplicatedMonitoringSubFunctionException,
            InvalidMonitoringObjectNameException, IllegalArgumentException,
            InvalidMonitoringSubFunctionNameException {

        if (monitoringObject != null) {

            final String monitoringObjectName = monitoringObject.getName();
            if (monitoringObjectName != null) {

                // We check that monitoring sub-functions are valids
                final List<MonitoringSubFunction> monitoringSubFunctions = monitoringObject
                        .getSubFunctions();
                if (monitoringSubFunctions != null) {
                    if (!monitoringSubFunctions.isEmpty()) {
                        final List<String> subFunctionNames = new ArrayList<>(monitoringSubFunctions.size());
                        for (final MonitoringSubFunction subFunction : monitoringSubFunctions) {
                            final String monitoringSubFunctionName = subFunction.getName();
                            if (monitoringSubFunctionName != null) {
                                if (subFunctionNames.contains(monitoringSubFunctionName)) {
                                    throw new DuplicatedMonitoringSubFunctionException(subFunction,
                                            monitoringObject);
                                } else {
                                    subFunctionNames.add(monitoringSubFunctionName);
                                }

                            } else {
                                throw new InvalidMonitoringSubFunctionNameException(
                                        monitoringObject, subFunction);
                            }
                        }
                    } else {
                        LOG.warning("Registration of a monitoring object ('"
                                + monitoringObjectName
                                + "') without sub-functions. Registration skipped and monitoring object not available !!");
                    }
                } else {
                    throw new IllegalArgumentException(
                            "The sub-function list of the monitoring object parameter is null");
                }

                // Sub-functions are valid, we can register the monitoring
                // object
                if (this.monitoringObjects.containsKey(monitoringObjectName)) {
                    throw new DuplicatedMonitoringObjectException(monitoringObject);
                } else {
                    this.monitoringObjects.put(monitoringObjectName, monitoringObject);
                }

            } else {
                throw new InvalidMonitoringObjectNameException(monitoringObject);
            }
        } else {
            throw new IllegalArgumentException("The monitoring object parameter is null");
        }
    }

    @Override
    public void registersSubscriptionObjects(final SubscriptionObject subscriptionObject)
            throws DuplicatedSubscriptionObjectException, DuplicatedEmittableDefectException,
            InvalidSubscriptionObjectNameException, InvalidEmittableDefectNameException,
            IllegalArgumentException {

        if (subscriptionObject != null) {

            final String subscriptionObjectName = subscriptionObject.getName();
            if (subscriptionObjectName != null) {

                // We check that emittable defect are valids
                final List<EmittableDefect> emittableDefects = subscriptionObject
                        .getEmittableDefects();
                if (emittableDefects != null) {
                    if (!emittableDefects.isEmpty()) {
                        final List<String> emittableDefectNames = new ArrayList<>(emittableDefects.size());
                        for (final EmittableDefect emittableDefect : emittableDefects) {
                            final String emittableDefectName = emittableDefect.getName();
                            if (emittableDefectName != null) {
                                if (emittableDefectNames.contains(emittableDefectName)) {
                                    throw new DuplicatedEmittableDefectException(emittableDefect,
                                            subscriptionObject);
                                } else {
                                    emittableDefectNames.add(emittableDefectName);
                                }

                            } else {
                                throw new InvalidEmittableDefectNameException(subscriptionObject,
                                        emittableDefect);
                            }
                        }
                    } else {
                        LOG.warning("Registration of a subscription object ('"
                                + subscriptionObjectName
                                + "') without emittable defect. Registration skipped and subscription object not available !!");
                    }
                } else {
                    throw new IllegalArgumentException(
                            "The emittable defect list of the subscription object parameter is null");
                }

                // Emittable defects are valid, we can register the subscription
                // object
                if (this.subscriptionObjects.containsKey(subscriptionObjectName)) {
                    throw new DuplicatedSubscriptionObjectException(subscriptionObject);
                } else {
                    this.subscriptionObjects.put(subscriptionObjectName, subscriptionObject);
                }

            } else {
                throw new InvalidSubscriptionObjectNameException(subscriptionObject);
            }
        } else {
            throw new IllegalArgumentException("The subscription object parameter is null");
        }
    }

    @Override
    public Options getOptions() {
        return this.options;
    }

    @Override
    public void resetOptions() {
        this.createOptions();
    }

    @Override
    public String getOptionsDescription() {
        final StringBuilder optsDescription = new StringBuilder();
        optsDescription.append(CommandUtil.formatCommandOptionsDescription(this));
        optsDescription.append('\n');
        optsDescription.append("MONITORING OBJECTS & SUB-FUNCTIONS");
        optsDescription.append('\n');

        for (final MonitoringObject monitoringObject : this.getMonitoringObjects().values()) {
            optsDescription.append("\t- ");
            optsDescription.append(monitoringObject.getName());
            optsDescription.append(": ");
            optsDescription.append(monitoringObject.getDescription());
            optsDescription.append('\n');

            for (final MonitoringSubFunction subFunction : monitoringObject.getSubFunctions()) {
                optsDescription.append("\t\t- ");
                optsDescription.append(subFunction.getName());
                optsDescription.append(": ");
                optsDescription.append(subFunction.getDescription());
                optsDescription.append('\n');
            }
        }

        optsDescription.append('\n');
        optsDescription.append("SUBSCRIPTION OBJECTS & DEFECTS");
        optsDescription.append('\n');

        for (final SubscriptionObject subscriptionObject : this.getSubscriptionObjects().values()) {
            optsDescription.append("\t- ");
            optsDescription.append(subscriptionObject.getName());
            optsDescription.append(": ");
            optsDescription.append(subscriptionObject.getDescription());
            optsDescription.append('\n');

            for (final EmittableDefect emittableDefect : subscriptionObject.getEmittableDefects()) {
                optsDescription.append("\t\t- ");
                optsDescription.append(emittableDefect.getName());
                optsDescription.append(": ");
                optsDescription.append(emittableDefect.getDescription());
                optsDescription.append('\n');
            }
        }

        return optsDescription.toString();
    }

    @Override
    public String getName() {
        return NAME;
    }

    @Override
    public String getUsage() {
        return CommandUtil.formatCommandUsage(this);
    }

    @Override
    public String getDescription() {
        return DESCRIPTION;
    }

    @Override
    public Completer getDefaultCompleter() {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Map<String, Completer> getOptionCompleters() {
        return this.completers;
    }

    @Override
    public Shell getShell() {
        return this.shell;
    }

    @Override
    public void setShell(final Shell shell) {
        this.shell = shell;
        for (final MonitoringObject monitoringObject : this.monitoringObjects.values()) {
            for (final MonitoringSubFunction monitoringSubFunction : monitoringObject
                    .getSubFunctions()) {
                monitoringSubFunction.setShell(shell);
            }
        }

        for (final SubscriptionObject subscriptionObject : this.subscriptionObjects.values()) {
            subscriptionObject.setShell(shell);
        }
    }

    private void createOptions() {

        this.options = new Options();
        this.options.addOption(OBJECT_OPTION);
        this.options.addOption(SUBFONCTION_OPTION);
        final OptionGroup subscribingGroup = new OptionGroup();
        subscribingGroup.addOption(SUBSCRIPTION_OPTION);
        subscribingGroup.addOption(UNSUBSCRIPTION_OPTION);
        this.options.addOptionGroup(subscribingGroup);
        this.options.addOption(HELP_OPTION);
    }

    @Override
    public Map<String, MonitoringObject> getMonitoringObjects() {
        return Collections.unmodifiableMap(this.monitoringObjects);
    }

    @Override
    public Map<String, SubscriptionObject> getSubscriptionObjects() {
        return Collections.unmodifiableMap(this.subscriptionObjects);
    }

    @Override
    public void execute(final String[] args) throws CommandException {
        
        if (this.isConnectionRequired() && !this.shell.isConnectionEstablished()) {
            throw new ConnectionRequiredException(this);
        }
        
        final CommandLineParser parser = new DefaultParser();
        try {
            final CommandLine cli = parser.parse(this.getOptions(), args);

            if (cli.hasOption(OBJECT_SHORT_OPTION) && cli.hasOption(SUBFONCTION_SHORT_OPTION)
                    && !cli.hasOption(SUBSCRIPTION_SHORT_OPTION)
                    && !cli.hasOption(UNSUBSCRIPTION_SHORT_OPTION)) {
                final String monitoringObjectName = cli.getOptionValue(OBJECT_SHORT_OPTION);
                final String monitoringSubFunctionName = cli
                        .getOptionValue(SUBFONCTION_SHORT_OPTION);
                final MonitoringObject monitoringObject = this.monitoringObjects
                        .get(monitoringObjectName);

                if (monitoringObject != null) {

                    // Retrieve the sub-function
                    MonitoringSubFunction monitoringSubFunctionFound = null;
                    for (final MonitoringSubFunction monitoringSubFunction : monitoringObject
                            .getSubFunctions()) {
                        if (monitoringSubFunction.getName().equals(monitoringSubFunctionName)) {
                            monitoringSubFunctionFound = monitoringSubFunction;
                        }
                    }

                    if (monitoringSubFunctionFound != null) {
                        if (cli.hasOption(HELP_SHORT_OPTION)) {
                            // Help on a monitoring object sub-function
                            this.getShell().getPrintStream()
                                    .println(getHelp(monitoringObject, monitoringSubFunctionName));
                        } else {
                            try {

                                monitoringSubFunctionFound.execute(cli.getArgs());

                            } catch (final MonitoringSubFunctionBadArgumentNumberException e) {
                                // Error on arguments, print help on the monitoring object sub-function
                                this.getShell().getErrorStream().println("ERROR: " + e.getMessage());
                                this.getShell().getPrintStream()
                                        .println(getUsage(monitoringObject, monitoringSubFunctionName));
                            } finally {
                                monitoringSubFunctionFound.resetOptions();
                            }
                        }
                    } else {
                        // Invalid monitoring sub-function
                        throw new CommandInvalidArgumentException(this, SUBFONCTION_OPTION,
                                monitoringSubFunctionName);
                    }
                } else {
                    // Invalid monitoring object
                    throw new CommandInvalidArgumentException(this, OBJECT_OPTION,
                            monitoringObjectName);
                }
            } else if (cli.hasOption(SUBSCRIPTION_SHORT_OPTION)
                    && !cli.hasOption(OBJECT_SHORT_OPTION)
                    && !cli.hasOption(SUBFONCTION_SHORT_OPTION)) {
                final String subscriptionObjectName = cli.getOptionValue(SUBSCRIPTION_SHORT_OPTION);
                final SubscriptionObject subscriptionObject = this.subscriptionObjects
                        .get(subscriptionObjectName);

                if (subscriptionObject != null) {
                    
                    if (cli.hasOption(HELP_SHORT_OPTION)) {
                        // Help on a monitoring object sub-function
                        this.getShell().getPrintStream().println(getHelp(subscriptionObject));
                    } else {
                        if (currentSubscriptions.contains(subscriptionObject.getName())) {
                            throw new SubscriptionAlreadyExistsException(this, subscriptionObject);
                        } else {
                            try {
                                subscriptionObject.subscribe(cli.getArgs());
                                currentSubscriptions.add(subscriptionObject.getName());
                            } finally {
                                subscriptionObject.resetOptions();
                            }
                        }
                    }

                } else {
                    // Invalid subscription object
                    throw new CommandInvalidArgumentException(this, SUBSCRIPTION_OPTION,
                            subscriptionObjectName);
                }
            } else if (cli.hasOption(UNSUBSCRIPTION_SHORT_OPTION)
                    && !cli.hasOption(OBJECT_SHORT_OPTION)
                    && !cli.hasOption(SUBFONCTION_SHORT_OPTION)) {
                final String subscriptionObjectName = cli
                        .getOptionValue(UNSUBSCRIPTION_SHORT_OPTION);
                final SubscriptionObject subscriptionObject = this.subscriptionObjects
                        .get(subscriptionObjectName);

                if (subscriptionObject != null) {
                    
                    if (cli.hasOption(HELP_SHORT_OPTION)) {
                        // Help on a monitoring object sub-function
                        this.getShell().getPrintStream().println(getHelp(subscriptionObject));
                    } else {
                        if (!currentSubscriptions.contains(subscriptionObject.getName())) {
                            throw new SubscriptionObjectNotSubscribedException(this,
                                    subscriptionObject);
                        } else {
                            try {
                                subscriptionObject.unsubscribe(cli.getArgs());
                                this.currentSubscriptions.remove(subscriptionObject.getName());
                            } finally {
                                subscriptionObject.resetOptions();
                            }
                        }
                    }
                    
                } else {
                    // Invalid subscription object
                    throw new CommandInvalidArgumentException(this, UNSUBSCRIPTION_OPTION,
                            subscriptionObjectName);
                }
            }
            else {
                throw new CommandBadArgumentNumberException(this);
            }
        } catch (final MonitoringSubFunctionException e) {
            throw new CommandException(this, e);
        } catch (final SubscriptionObjectException e) {
            throw new CommandException(this, e);
        } catch (final UnrecognizedOptionException e) {
            throw new CommandBadArgumentNumberException(this, e);
        } catch (final MissingArgumentException e) {
            throw new CommandMissingArgumentException(this, e.getOption(), e);
        } catch (final MissingOptionException e) {
            throw new CommandMissingOptionsException(this, e.getMissingOptions(), e);
        } catch (final ParseException e) {
            throw new CommandInvalidException(this, e);
        }

    }

    private static String getHelp(final MonitoringObject monitoringObject,
            final String monitoringSubFunctionName) {

        for (final MonitoringSubFunction monitoringSubFunction : monitoringObject.getSubFunctions()) {
            if (monitoringSubFunction.getName().equals(monitoringSubFunctionName)) {

                final Options subFunctionOptions = monitoringSubFunction.getOptions();

                final String subfctDescr = monitoringSubFunction.getDescription();
                if (subFunctionOptions == null) {
                    if (subfctDescr != null) {
                        return String
                                .format("%nUSAGE:%n MONITORING OBJECT: %s%n SUB-FUNCTION: %s%n%nDESCRIPTION:%n %s%n%n",
                                        monitoringObject.getName(), monitoringSubFunctionName,
                                        monitoringSubFunction.getDescription());
                    } else {
                        return String.format(
                                "%nUSAGE:%n MONITORING OBJECT: %s%n SUB-FUNCTION: %s%n%n",
                                monitoringObject.getName(), monitoringSubFunctionName);

                    }
                } else {
                    // use help formatter to provide well-formed description
                    final HelpFormatter usageFormatter = new HelpFormatter();
                    usageFormatter.setSyntaxPrefix("");
                    final StringWriter usageSW = new StringWriter();
                    usageFormatter.printUsage(new PrintWriter(usageSW),
                            HelpFormatter.DEFAULT_WIDTH, "", monitoringSubFunction.getOptions());
                    // use help formatter to provide well-formed description
                    final StringWriter optionsDescriptionSW = new StringWriter();
                    final HelpFormatter helpFormatter = new HelpFormatter();
                    helpFormatter.printOptions(new PrintWriter(optionsDescriptionSW),
 HelpFormatter.DEFAULT_WIDTH,
                            subFunctionOptions, HelpFormatter.DEFAULT_LEFT_PAD,
                            HelpFormatter.DEFAULT_DESC_PAD);



                    if (subfctDescr != null) {
                        return String
                                .format("%nUSAGE:%n MONITORING OBJECT: %s%n SUB-FUNCTION: %s%n%n%s%n%nDESCRIPTION:%n %s%n%nOPTIONS DESCRIPTION:%n%s%n%n",
                                        monitoringObject.getName(), monitoringSubFunctionName,
                                        usageSW.toString(), monitoringSubFunction.getDescription(),
                                        optionsDescriptionSW.toString());
                    } else {
                        return String
                                .format("%nUSAGE:%n MONITORING OBJECT: %s%n SUB-FUNCTION: %s%n%n%s%n%nOPTIONS DESCRIPTION:%n%s%n%n",
                                        monitoringObject.getName(), monitoringSubFunctionName,
                                        usageSW.toString(), optionsDescriptionSW.toString());
                    }
                }
            }
        }

        // Here, no sub-function found. This should not occur because already
        // processed by the method 'execute()'
        return null;
    }

    private static String getHelp(final SubscriptionObject subscriptionObject) {
        
        final Options subFunctionOptions = subscriptionObject.getOptions();

        final StringBuilder defectsDescription = new StringBuilder();
        for (final EmittableDefect emittableDefect : subscriptionObject.getEmittableDefects()) {
            final String defectDescr = emittableDefect.getDescription();
            if (defectDescr != null) {
                defectsDescription.append(String.format(" - %s: %s%n", emittableDefect.getName(),
                        defectDescr));
            } else {
                defectsDescription.append(String.format(" - %s%n", emittableDefect.getName()));
            }
        }

        final String soDescr = subscriptionObject.getDescription();
        if (subFunctionOptions == null) {
            if (soDescr != null) {
                return String.format(
                        "%nUSAGE:%n SUBSCRIPTION OBJECT: %s%n%nDESCRIPTION:%n %s%n%nDEFECT:%n%s%n",
                        subscriptionObject.getName(), subscriptionObject.getDescription(),
                        defectsDescription);
            } else {
                return String.format("%nUSAGE:%n SUBSCRIPTION OBJECT: %s%n%nDEFECT:%n%s%n",
                        subscriptionObject.getName(), defectsDescription);

            }
        } else {
            // use help formatter to provide well-formed description
            final HelpFormatter usageFormatter = new HelpFormatter();
            usageFormatter.setSyntaxPrefix("");
            final StringWriter usageSW = new StringWriter();
            usageFormatter.printUsage(new PrintWriter(usageSW), HelpFormatter.DEFAULT_WIDTH, "",
 subFunctionOptions);
            // use help formatter to provide well-formed description
            final StringWriter optionsDescriptionSW = new StringWriter();
            final HelpFormatter helpFormatter = new HelpFormatter();
            helpFormatter.printOptions(new PrintWriter(optionsDescriptionSW),
 HelpFormatter.DEFAULT_WIDTH,
                    subFunctionOptions, HelpFormatter.DEFAULT_LEFT_PAD,
                    HelpFormatter.DEFAULT_DESC_PAD);

            if (soDescr != null) {
                return String
                        .format("%nUSAGE:%n SUBSCRIPTION OBJECT: %s%n%n%s%n%nDESCRIPTION:%n %s%n%nOPTIONS DESCRIPTION:%n%s%n%nDEFECT:%n%s%n",
                                subscriptionObject.getName(), usageSW.toString(),
                                subscriptionObject.getDescription(),
                                optionsDescriptionSW.toString(), defectsDescription);
            } else {
                return String
                        .format("%nUSAGE:%n SUBSCRIPTION OBJECT: %s%n%n%s%n%nOPTIONS DESCRIPTION:%n%s%n%nDEFECT:%n%s%n",
                                subscriptionObject.getName(), usageSW.toString(),
                                optionsDescriptionSW.toString(), defectsDescription);
            }
        }

    }

    private static String getUsage(final MonitoringObject monitoringObject,
            final String monitoringSubFunctionName) {

        for (final MonitoringSubFunction monitoringSubFunction : monitoringObject.getSubFunctions()) {
            if (monitoringSubFunction.getName().equals(monitoringSubFunctionName)) {

                // use help formatter to provide well-formed description
                final HelpFormatter usageFormatter = new HelpFormatter();
                usageFormatter.setSyntaxPrefix("");
                final StringWriter usageSW = new StringWriter();
                usageFormatter.printUsage(new PrintWriter(usageSW), HelpFormatter.DEFAULT_WIDTH,
                        "", monitoringSubFunction.getOptions());

                return usageSW.toString();
            }
        }

        // Here, no sub-function found. This should not occur because already
        // processed by the method 'execute()'
        return null;
    }

    /**
     * {@inheritDoc}
     * 
     * <p>
     * Dedicated processing to the command 'monitoring': a connection is
     * required.
     * </p>
     */
    @Override
    public boolean isConnectionRequired() {
        return true;
    }

}
